/*         ______   ___    ___
 *        /\  _  \ /\_ \  /\_ \
 *        \ \ \L\ \\//\ \ \//\ \      __     __   _ __   ___ 
 *         \ \  __ \ \ \ \  \ \ \   /'__`\ /'_ `\/\`'__\/ __`\
 *          \ \ \/\ \ \_\ \_ \_\ \_/\  __//\ \L\ \ \ \//\ \L\ \
 *           \ \_\ \_\/\____\/\____\ \____\ \____ \ \_\\ \____/
 *            \/_/\/_/\/____/\/____/\/____/\/___L\ \/_/ \/___/
 *                                           /\____/
 *                                           \_/__/
 *
 *      File I/O subsystem
 *
 *      By Julian Smythe.
 *
 *      See readme.txt for copyright information.
 */

#include "pcommon.h"

int _al_file_isok(AL_CONST char *filename)
{
	return TRUE;
}

uint64_t _al_file_size_ex(AL_CONST char *filename)
{
	struct stat s;
	char tmp[1024];
	
	if (stat(uconvert_toascii(filename,tmp), &s) != 0)
	{
		*allegro_errno = errno;
		return 0;
	}

	return s.st_size;
}

time_t _al_file_time(AL_CONST char *filename)
{
	struct stat s;
	char tmp[1024];
	
	if (stat(uconvert_toascii(filename,tmp),&s) != 0)
	{
		*allegro_errno = errno;
		return 0;
	}

	return s.st_mtime;
}

static void ff_put_backslash(char *filename, int size)
{
	int len = strlen(filename);

	if (len > 0 && len < (size-1) && filename[len-1] != '/')
	{
		filename[len] = '/';
		filename[len+1] = 0;
	}
}

#define FF_MATCH_TRY 0
#define FF_MATCH_ONE 1
#define FF_MATCH_ANY 2

struct FF_MATCH_DATA
{
	int type;
	AL_CONST char *s1;
	AL_CONST char *s2;
};

static int ff_match(AL_CONST char *s1, AL_CONST char *s2)
{
	static unsigned int size = 0;
	static struct FF_MATCH_DATA *data = NULL;
	AL_CONST char *s1end;
	int index, c1, c2;

	if ((!s1) && (!s2))
	{
		if (data)
		{
			_AL_FREE(data);
			data = NULL;
		}

		return 0;
	}

	s1end = s1 + strlen(s1);

	if (data && size < strlen(s2))
	{
		_AL_FREE(data);
		data = NULL;
	}

	if (!data)
	{
		size = strlen(s2);
		data = _AL_MALLOC(sizeof(struct FF_MATCH_DATA)*size*2+1);
		if (!data)
		{
			return 0;
		}
	}

	index = 0;
	data[0].s1 = s1;
	data[0].s2 = s2;
	data[0].type = FF_MATCH_TRY;

	while (index >= 0)
	{
		s1 = data[index].s1;
		s2 = data[index].s2;
		c1 = *s1;
		c2 = *s2;

		switch (data[index].type)
		{

		case FF_MATCH_TRY:
			if (c2 == 0)
			{
				if (c1 == 0)
					return 1;
				else
					index--;
			}
			else if (c1 == 0)
			{
				while (*s2 == '*')
					s2++;
				if (*s2 == 0)
					return 1;
				else
					index--;
			}
			else if (c2 == '*')
			{
				data[index++].type = FF_MATCH_ANY;
				data[index].s1 = s1end;
				data[index].s2 = s2+1;
				data[index].type = FF_MATCH_TRY;
			}
			else if ((c2 == '?') || (c1 == c2))
			{
				data[index++].type = FF_MATCH_ONE;
				data[index].s1 = s1+1;
				data[index].s2 = s2+1;
				data[index].type = FF_MATCH_TRY;
			}
			else
			{
				index--;
			}
			break;

		case FF_MATCH_ONE:
			index--;
			break;

		case FF_MATCH_ANY:
			if (--data[index+1].s1 >= s1)
			{
				data[index+1].type = FF_MATCH_TRY;
				index++;
			}
			else
			{
				index--;
			}
			break;

		default:
			return 0;
		}
	}

	return 0;
}

static int ff_get_attrib(AL_CONST char *name, struct stat *s)
{
	int attrib = 0;

	if (S_ISDIR(s->st_mode))
	{
		attrib |= FA_DIREC;
	}
 
	if (name[0] == '.' && (name[1] != '.' || name[2] != '\0'))
	{
		attrib |= FA_HIDDEN;
	}

	return attrib;
}

struct FF_DATA
{
	SceUID dir;
	char dirname[1024];
	char pattern[1024];
	int attrib;
	uint64_t size;
};

static char *ff_get_filename(AL_CONST char *path)
{
	char *p = (char *)path + strlen(path);

	while (p > path && p[-1] != '/')
	{
		--p;
	}

	return p;
}


int al_findfirst(AL_CONST char *pattern, struct al_ffblk *info, int attrib)
{
	struct FF_DATA *ff_data;
	struct stat s;
	int actual_attrib;
	char tmp[1024];
	char *p;

	ff_data = (struct FF_DATA *)_AL_MALLOC(sizeof(struct FF_DATA));
	if (!ff_data)
	{
		*allegro_errno = ENOMEM;
		return -1;
	}

	memset(ff_data,0,sizeof(*ff_data));
	info->ff_data = (void *)ff_data;

	if (!ustrpbrk(pattern,uconvert("?*",U_ASCII,tmp,U_CURRENT,sizeof(tmp))))
	{
		errno = *allegro_errno = 0;

		if (stat(uconvert(pattern,U_CURRENT,tmp,U_UTF8,sizeof(tmp)),&s) == 0)
		{
			actual_attrib = ff_get_attrib(ff_get_filename(uconvert(pattern,U_CURRENT,tmp,U_UTF8,sizeof(tmp))),&s);

			if ((actual_attrib & ~attrib) == 0)
			{
				info->attrib = actual_attrib;
				info->time = s.st_mtime;
				info->size = s.st_size;
				ff_data->size = s.st_size;
				ustrzcpy(info->name, sizeof(info->name),get_filename(pattern));
				return 0;
			}
		}

		_AL_FREE(ff_data);
		info->ff_data = NULL;
		*allegro_errno = errno ? errno : ENOENT;
		return -1;
	}

	ff_data->attrib = attrib;

	do_uconvert(pattern,U_CURRENT,ff_data->dirname,U_UTF8,sizeof(ff_data->dirname));
	p = ff_get_filename(ff_data->dirname);
	_al_sane_strncpy(ff_data->pattern,p,sizeof(ff_data->pattern));
	if (p == ff_data->dirname)
	{
		_al_sane_strncpy(ff_data->dirname,"./",1024);
	}
	else
	{
		*p = 0;
	}

	if (strcmp(ff_data->pattern,"*.*") == 0)
	{
		_al_sane_strncpy(ff_data->pattern,"*",1024);
	}

	errno = *allegro_errno = 0;

	ff_data->dir = sceIoDopen(ff_data->dirname);

	if (!ff_data->dir)
	{
		*allegro_errno = errno ? errno : ENOENT;
		_AL_FREE(ff_data);
		info->ff_data = NULL;
		return -1;
	}

	if (al_findnext(info) != 0)
	{
		al_findclose(info);
		return -1;
	}

	return 0;
}

int al_findnext(struct al_ffblk *info)
{
	char tempname[1024];
	char filename[1024];
	int attrib;
	struct SceIoDirent entry;
	struct stat s;
	struct FF_DATA *ff_data = (struct FF_DATA *) info->ff_data;

	ASSERT(ff_data);

	if (!ff_data->dir)
	{
		return -1;
	}
	
	while (1)
	{
		if (sceIoDread(ff_data->dir,&entry) < 0)
		{
			*allegro_errno = errno ? errno : ENOENT;
			return -1;
		}

		tempname[0] = 0;
		if (strlen(entry.d_name) >= sizeof(tempname))
		{
			strncat(tempname,entry.d_name,sizeof(tempname)-1);
		}
		else
		{
			strncat(tempname,entry.d_name,strlen(entry.d_name));
		}

		if (ff_match(tempname,ff_data->pattern))
		{
			_al_sane_strncpy(filename,ff_data->dirname,1024);
			ff_put_backslash(filename,sizeof(filename));
			strncat(filename,tempname,sizeof(filename)-strlen(filename)-1);

			if (stat(filename,&s) == 0)
			{
				attrib = ff_get_attrib(tempname,&s);

				if ((attrib & ~ff_data->attrib) == 0)
				{
					break;
				}
			}
			else
			{
				*allegro_errno = 0;
			}
		}
	}

	info->attrib = attrib;
	info->time = s.st_mtime;
	info->size = s.st_size;
	ff_data->size = s.st_size;

	do_uconvert(tempname,U_UTF8,info->name,U_CURRENT,sizeof(info->name));

	return 0;
}

void al_findclose(struct al_ffblk *info)
{
	struct FF_DATA *ff_data = (struct FF_DATA *)info->ff_data;

	if (ff_data)
	{
		if (ff_data->dir)
		{
			sceIoDclose(ff_data->dir);
		}

		_AL_FREE(ff_data);
		info->ff_data = NULL;
		ff_match(NULL, NULL);
	}
}

void _al_getdcwd(int drive, char *buf, int size)
{
	char tmp[1024];

	if (getcwd(tmp, sizeof(tmp)))
	{
		do_uconvert(tmp,U_UTF8,buf,U_CURRENT,size);
	}
	else
	{
		usetc(buf,0);
	}
}

