/*         ______   ___    ___
 *        /\  _  \ /\_ \  /\_ \
 *        \ \ \L\ \\//\ \ \//\ \      __     __   _ __   ___ 
 *         \ \  __ \ \ \ \  \ \ \   /'__`\ /'_ `\/\`'__\/ __`\
 *          \ \ \/\ \ \_\ \_ \_\ \_/\  __//\ \L\ \ \ \//\ \L\ \
 *           \ \_\ \_\/\____\/\____\ \____\ \____ \ \_\\ \____/
 *            \/_/\/_/\/____/\/____/\/____/\/___L\ \/_/ \/___/
 *                                           /\____/
 *                                           \_/__/
 *
 *      Graphics driver
 *
 *      By Julian Smythe.
 *
 *      See readme.txt for copyright information.
 */

#include "pcommon.h"

// Buffer to store GU commands
// 32Ko of memory should be more than enough
static unsigned char PSP_ALIGNED(16) gfx_psp_display_list[32*1024];

static unsigned long gfx_psp_color_depth_const(int color_depth)
{
	switch (color_depth)
	{
		case 8: case 15: case 16: return GU_PSM_5650;
		case 24: case 32: return GU_PSM_8888;
		default: return 0;
	}
}

static void *gfx_psp_fbp0;
static void *gfx_psp_zbp;

struct BITMAP *gfx_psp_init(int w, int h, int v_w, int v_h, int color_depth);
void gfx_psp_exit(struct BITMAP *b);
void gfx_psp_vsync(void);

GFX_DRIVER gfxdrv_psp =
{
	GFX_PSP,
	"GFX_PSP",
	"PSP Graphics Driver",
	"GFX_PSP",
	gfx_psp_init,
	gfx_psp_exit,
	NULL, //gfx_psp_scroll
	gfx_psp_vsync,
	NULL, //gfx_psp_set_palette
	NULL, //gfx_psp_request_scroll
	NULL, //gfx_psp_poll_scroll
	NULL, //gfx_psp_enable_triple_buffer
	NULL, //gfx_psp_create_video_bitmap
	NULL, //gfx_psp_destroy_video_bitmap
	NULL, //gfx_psp_show_video_bitmap
	NULL, //gfx_psp_request_video_bitmap
	NULL, //gfx_psp_create_system_bitmap
	NULL, //gfx_psp_destroy_system_bitmap
	NULL, //gfx_psp_set_mouse_sprite
	NULL, //gfx_psp_show_mouse
	NULL, //gfx_psp_hide_mouse
	NULL, //gfx_psp_move_mouse
	NULL, //gfx_psp_drawing_mode
	NULL, //gfx_psp_save_video_state
	NULL, //gfx_psp_restore_video_state
	NULL, //gfx_psp_set_blender_mode
	NULL, //gfx_psp_fetch_mode_list
	PSP_SCREEN_W,
	PSP_SCREEN_H,
	TRUE,
	0,
	0,
	0, // sceGeEdramGetSize()
	0, // (0x40000000 | (unsigned long)sceGeEdramGetAddr())
	FALSE
};

struct BITMAP *gfx_psp_init(int w, int h, int v_w, int v_h, int color_depth)
{
	BITMAP *psp_screen;
	int bpp;
	
	if (color_depth == 8 || color_depth == 15 || color_depth == 16)
	{
		set_color_depth(32);
		color_depth = 32;
	}
	
	bpp = color_depth/8;
	
	TRACE(" gfx_psp_init was called\n");
	
	gfxdrv_psp.vid_mem = sceGeEdramGetSize();
	gfxdrv_psp.vid_phys_base = (0x40000000 | (unsigned long)sceGeEdramGetAddr());
	
	// These pointers are relative to VRAM, not the main memory
	gfx_psp_fbp0 = (void *)0;
	gfx_psp_zbp = (void *)(gfx_psp_fbp0 + PSP_A_SCREEN_W*PSP_SCREEN_H*bpp);

	sceGuInit();

	sceGuStart(GU_DIRECT,gfx_psp_display_list);
	sceGuDrawBuffer(gfx_psp_color_depth_const(color_depth),gfx_psp_fbp0,PSP_A_SCREEN_W);
	sceGuDispBuffer(w,h,gfx_psp_fbp0,PSP_A_SCREEN_W);
	sceGuDepthBuffer(gfx_psp_zbp,PSP_A_SCREEN_W);
	sceGuOffset(2048 - (w/2),2048 - (h/2));
	sceGuViewport(2048,2048,w,h);
	sceGuDepthRange(65535,0);
	sceGuScissor(0,0,w,h);
	sceGuEnable(GU_SCISSOR_TEST);
	sceGuFrontFace(GU_CW);
	sceGuEnable(GU_TEXTURE_2D);
	sceGuClear(GU_COLOR_BUFFER_BIT|GU_DEPTH_BUFFER_BIT);
	sceGuFinish();

	sceGuSync(0,0);
	sceDisplayWaitVblankStart();
	sceGuDisplay(1);

	psp_screen = _make_bitmap(w,h,(unsigned int)(gfxdrv_psp.vid_phys_base+gfx_psp_fbp0),&gfxdrv_psp,color_depth,PSP_A_SCREEN_W*bpp);
	gfxdrv_psp.w = w;
	gfxdrv_psp.h = h;
	
	if (psp_screen == NULL)
	{
		TRACE(" _make_bitmap FAILED\n");
	}

	return psp_screen;
}

void gfx_psp_exit(struct BITMAP *b)
{
	_AL_FREE(b);
}

void gfx_psp_vsync(void)
{
	sceDisplayWaitVblankStart();
}
